@@ -61,6 +61,7 @@ class CqlSessionProvider(object):
         self._auth_provider = None
         self._ssl_context = None
         self._cassandra_config = cassandra_config
+        self._native_port = CassandraConfigReader(cassandra_config.config_file).native_port
 
         if null_if_empty(cassandra_config.cql_username) and null_if_empty(cassandra_config.cql_password):
             auth_provider = PlainTextAuthProvider(username=cassandra_config.cql_username,
@@ -90,6 +91,7 @@ class CqlSessionProvider(object):
          """
 
         cluster = Cluster(contact_points=self._ip_addresses,
+                          port=int(self._native_port),
                           auth_provider=self._auth_provider,
                           execution_profiles=self._execution_profiles,
                           ssl_context=self._ssl_context)
@@ -147,8 +149,8 @@ class CqlSession(object):
                 return token.value
         raise RuntimeError('Unable to get current token')
 
-    def datacenter(self):
-        logging.debug('Checking datacenter...')
+    def placement(self):
+        logging.debug('Checking placement using dc and rack...')
         listen_address = socket.gethostbyname(self.cluster.contact_points[0])
         token_map = self.cluster.metadata.token_map
 
@@ -156,13 +158,13 @@ class CqlSession(object):
             socket_host = self.hostname_resolver.resolve_fqdn(listen_address)
             logging.debug('Checking host {} against {}/{}'.format(host.address, listen_address, socket_host))
             if host.address == listen_address or self.hostname_resolver.resolve_fqdn(host.address) == socket_host:
-                return host.datacenter
+                return host.datacenter, host.rack
 
-        raise RuntimeError('Unable to get current datacenter')
+        raise RuntimeError('Unable to get current placement')
 
     def tokenmap(self):
         token_map = self.cluster.metadata.token_map
-        datacenter = self.datacenter()
+        dc_rack_pair = self.placement()
 
         def get_host(host_token_pair):
             return host_token_pair[0]
@@ -183,10 +185,12 @@ class CqlSession(object):
         return {
             self.hostname_resolver.resolve_fqdn(host.address): {
                 'tokens': tokens,
-                'is_up': host.is_up
+                'is_up': host.is_up,
+                'rack': host.rack,
+                'dc': host.datacenter
             }
             for host, tokens in host_tokens_pairs
-            if host.datacenter == datacenter
+            if host.datacenter == dc_rack_pair[0]
         }
 
     def dump_schema(self):
@@ -325,19 +329,17 @@ class Cassandra(object):
         self._is_ccm = int(shlex.split(cassandra_config.is_ccm)[0])
         self._os_has_systemd = self._has_systemd()
         self._nodetool = Nodetool(cassandra_config)
-        logging.warning('is ccm : {}'.format(self._is_ccm))
         config_reader = CassandraConfigReader(cassandra_config.config_file, release_version)
         self._cassandra_config_file = cassandra_config.config_file
         self._root = config_reader.root
         self._commitlog_path = config_reader.commitlog_directory
         self._saved_caches_path = config_reader.saved_caches_directory
         self._hostname = contact_point if contact_point is not None else config_reader.listen_address
+        self._storage_port = config_reader.storage_port
+        self._native_port = config_reader.native_port
         self._cql_session_provider = CqlSessionProvider(
             [self._hostname],
             cassandra_config)
-
-        self._storage_port = config_reader.storage_port
-        self._native_port = config_reader.native_port
         self._rpc_port = config_reader.rpc_port
         self.seeds = config_reader.seeds
 
@@ -707,8 +709,14 @@ def is_open(host, port):
         s.connect((host, port))
         s.shutdown(socket.SHUT_RDWR)
         is_accessible = True
+
+    # If cassandra is not running but the host is up, host may choose to silently drop inbound connections to the
+    #   closed port or may respond with a RST indicating that the connection was refused.
+    # ConnectionRefusedError: [Errno 111] Connection refused
+    except ConnectionRefusedError as cre:
+        logging.error("Host '{host}' is up, but port '{port}' is closed.".format(host=host, port=port), exc_info=cre)
     except socket.error as e:
-        logging.error('Port {} closed on host {}'.format(port, host), exc_info=e)
+        logging.error("Could not open socket to port '{port}' on '{host}'.".format(host=host, port=port), exc_info=e)
     finally:
         try:
             if s:
