@@ -43,7 +43,7 @@ def main(config, max_backup_age=0, max_backup_count=0):
         # list all backups to purge based on count conditions
         backups_to_purge |= set(backups_to_purge_by_count(backups, max_backup_count))
         # purge all candidate backups
-        purge_backups(storage, backups_to_purge)
+        purge_backups(storage, backups_to_purge, config.storage.backup_grace_period_in_days, config.storage.fqdn)
 
         logging.debug('Emitting metrics')
         tags = ['medusa-node-backup', 'purge-error', 'PURGE-ERROR']
@@ -83,7 +83,7 @@ def backups_to_purge_by_count(backups, max_backup_count):
 #     return list(filter(lambda backup: backup.name = backup_name, backups)) or list()
 
 
-def purge_backups(storage, backups):
+def purge_backups(storage, backups, backup_grace_period_in_days, local_fqdn):
     """
     Core function to purge a set of node_backups
     Used for node purge and backup delete (using a specific backup_name)
@@ -92,6 +92,7 @@ def purge_backups(storage, backups):
     fqdns = set()
     nb_objects_purged = 0
     total_purged_size = 0
+    total_objects_within_grace = 0
 
     for backup in backups:
         (purged_objects, purged_size) = purge_backup(storage, backup)
@@ -99,15 +100,27 @@ def purge_backups(storage, backups):
         total_purged_size += purged_size
         fqdns.add(backup.fqdn)
 
+    if len(fqdns) == 0:
+        # If we didn't purge any backup, we still want to cleanup obsolete files for the local node
+        fqdns.add(local_fqdn)
+
     for fqdn in fqdns:
-        (cleaned_objects_count, cleaned_objects_size) = cleanup_obsolete_files(storage, fqdn)
+        (cleaned_objects_count, cleaned_objects_size, nb_objects_within_grace) \
+            = cleanup_obsolete_files(storage,
+                                     fqdn,
+                                     backup_grace_period_in_days)
         nb_objects_purged += cleaned_objects_count
         total_purged_size += cleaned_objects_size
+        total_objects_within_grace += nb_objects_within_grace
 
     logging.info("Purged {} objects with a total size of {}".format(
         nb_objects_purged,
-        format_bytes_str(total_purged_size))
-    )
+        format_bytes_str(total_purged_size)))
+    if total_objects_within_grace > 0:
+        logging.info("{} objects within {} days grace period were not deleted".format(
+            total_objects_within_grace,
+            backup_grace_period_in_days
+        ))
 
 
 def purge_backup(storage, backup):
@@ -128,16 +141,21 @@ def purge_backup(storage, backup):
     return (purged_objects, purged_size)
 
 
-def cleanup_obsolete_files(storage, fqdn):
+def cleanup_obsolete_files(storage, fqdn, backup_grace_period_in_days):
     logging.info("Cleaning up orphaned files for {}...".format(fqdn))
     nb_objects_purged = 0
     total_purged_size = 0
 
     backups = storage.list_node_backups(fqdn=fqdn)
-    paths_in_manifest = get_file_paths_from_manifests_for_differential_backups(backups)
+    paths_in_manifest = get_file_paths_from_manifests_for_complete_differential_backups(backups)
     paths_in_storage = get_file_paths_from_storage(storage, fqdn)
 
-    for path in paths_in_storage - paths_in_manifest:
+    deletion_candidates = set(paths_in_storage.keys()) - paths_in_manifest
+    objects_to_delete = filter_files_within_gc_grace(storage,
+                                                     deletion_candidates,
+                                                     paths_in_storage,
+                                                     backup_grace_period_in_days)
+    for path in objects_to_delete:
         logging.debug("  - [{}] exists in storage, but not in manifest".format(path))
         obj = storage.storage_driver.get_blob(path)
         if obj is not None:
@@ -145,7 +163,9 @@ def cleanup_obsolete_files(storage, fqdn):
             total_purged_size += int(obj.size)
             storage.storage_driver.delete_object(obj)
 
-    return nb_objects_purged, total_purged_size
+    nb_objects_within_grace = len(set(deletion_candidates) - set(objects_to_delete))
+
+    return nb_objects_purged, total_purged_size, nb_objects_within_grace
 
 
 def get_file_paths_from_storage(storage, fqdn):
@@ -155,13 +175,26 @@ def get_file_paths_from_storage(storage, fqdn):
         for blob in storage.storage_driver.list_objects(str(data_directory))
     }
 
-    return set(data_files.keys())
+    return data_files
 
 
-def get_file_paths_from_manifests_for_differential_backups(backups):
+def filter_files_within_gc_grace(storage, blobs, paths_in_storage, backup_grace_period_in_days):
+    return list(
+        filter(lambda blob:
+               is_older_than_gc_grace(storage.storage_driver.get_object_datetime(paths_in_storage[blob]),
+                                      backup_grace_period_in_days),
+               blobs))
+
+
+def is_older_than_gc_grace(blob_datetime, gc_grace) -> bool:
+    return datetime.timestamp(blob_datetime) <= datetime.timestamp(datetime.now()) - (int(gc_grace) * 86400)
+
+
+def get_file_paths_from_manifests_for_complete_differential_backups(backups):
     differential_backups = filter_differential_backups(backups)
+    complete_differential_backups = list(filter(lambda backup: backup.manifest is not None, differential_backups))
 
-    manifests = list(map(lambda backup: json.loads(backup.manifest), differential_backups))
+    manifests = list(map(lambda backup: json.loads(backup.manifest), complete_differential_backups))
 
     objects_in_manifests = [
         obj
@@ -182,20 +215,40 @@ def filter_differential_backups(backups):
     return list(filter(lambda backup: backup.is_differential is True, backups))
 
 
-def delete_backup(config, backup_name, all_nodes):
+def backups_to_purge_by_name(storage, cluster_backups, backup_names_to_purge, all_nodes):
+    """
+    select NodeBackups that should be purged by backup name.
+    Raises KeyError exception if some of the backup_names_to_purge do not exist.
+    :param storage: storage object
+    :param cluster_backups: list of ClusterBackups for cluster
+    :param backup_names_to_purge: names of backups to purge
+    :param all_nodes: purge for all nodes if true, otherwise for current node only
+    :return: list of NodeBackups that should be purged
+    """
     backups_to_purge = list()
+    cluster_backups_by_name = {bk.name: bk for bk in cluster_backups}
+    for backup_name in backup_names_to_purge:
+        if backup_name in cluster_backups_by_name:
+            backups_to_purge.extend(cluster_backups_by_name[backup_name].node_backups.values())
+        else:
+            raise KeyError('The backup {} does not exist'.format(backup_name))
+
+    if not all_nodes:
+        backups_to_purge = [nb for nb in backups_to_purge if storage.config.fqdn in nb.fqdn]
+
+    return backups_to_purge
+
+
+def delete_backup(config, backup_names, all_nodes):
     monitoring = Monitoring(config=config.monitoring)
 
     try:
         storage = Storage(config=config.storage)
-        cluster_backup = storage.get_cluster_backup(backup_name)
-        backups_to_purge = cluster_backup.node_backups.values()
+        cluster_backups = storage.list_cluster_backups()
+        backups_to_purge = backups_to_purge_by_name(storage, cluster_backups, backup_names, all_nodes)
 
-        if not all_nodes:
-            backups_to_purge = [nb for nb in backups_to_purge if storage.config.fqdn in nb.fqdn]
-
-        logging.info('Deleting Backup {}...'.format(backup_name))
-        purge_backups(storage, backups_to_purge)
+        logging.info('Deleting Backup(s) {}...'.format(",".join(backup_names)))
+        purge_backups(storage, backups_to_purge, config.storage.backup_grace_period_in_days, storage.config.fqdn)
 
         logging.debug('Emitting metrics')
         tags = ['medusa-node-backup', 'delete-error', 'DELETE-ERROR']
@@ -205,6 +258,6 @@ def delete_backup(config, backup_name, all_nodes):
         monitoring.send(tags, 1)
         medusa.utils.handle_exception(
             e,
-            'This error happened during the delete of backup "{}": {}'.format(backup_name, str(e)),
+            'This error happened during the delete of backup(s) "{}": {}'.format(",".join(backup_names), str(e)),
             config
         )
